/**
 * CLASS: Beam
 * AUTHOR: danpost (greenfoot username)
 * CREATED: January 27, 2017
 * 
 * DESCRIPTION:  each beam created from this class is a straight-line variable length object
 * with possible source and target objects;
 */

import greenfoot.*;

public class Beam extends Actor
{
    private Beam priorBeam; // the beam this actor represent the reflected beam of, if applicable
    private Actor source, target; // the objects at either end of the beam (laser or reflector)
    private int origX, origY, maxWide, shortened; // fields to control length and position
    private boolean targetFound; // flag used to indicate beam ends
    private int delay;

    /**
     * used by the main source of the beam using:  new Beam(null, this)
     * 
     * @param src the object creating the beam
     */
    public Beam(Actor src)
    {
        this(null, src);
    }

    /**
     * creates a beam
     * 
     * @param pBeam the incident beem, if this is a reflected beam
     * @param src the object creating or reflecting this beam
     */
    public Beam(Beam pBeam, Actor src)
    {
        priorBeam = pBeam;
        source = src;
        maxWide = 2;
        // the initial image
        GreenfootImage image = new GreenfootImage(maxWide, 2);
        image.setColor(Color.YELLOW);
        image.fill();
        setImage(image);
    }
    
    /**
     * prepares the beam, setting its proper rotation and initial position in the world
     * 
     * @param world the world the beam was added into
     */
    protected void addedToWorld(World world)
    {
        // save initial position
        origX = getX();
        origY = getY();
        // set proper rotation
        if (priorBeam == null) setRotation(source.getRotation());
        else setRotation(2*source.getRotation()-priorBeam.getRotation());
        // added to prevent newly created from being removed when shortened by change in source
        lengthen();
    }

    /** controls the length of the beam while in world */
    public void act()
    {
        if (delay < 2 && ++delay < 2) return;
        // check for removal of beam that this represents the reflection of, if applicable
        if (priorBeam != null && priorBeam.getWorld() == null) source = null;
        // shorten unsourced beam to removal
        if (source == null && !shorten()) return;
        // lengthen beam while checking for an end destination (edge of world or obstacle)
        if (!targetFound)
        {
            lengthen();
             // perform checks at end of beam
            toEnd(); // position at end of beam
            // edge check
            boolean reachedXEdge = getX() == 0 || getX() == getWorld().getWidth()-1;
            boolean reachedYEdge =  getY() == 0 || getY() == getWorld().getHeight()-1;
            if (reachedXEdge || reachedYEdge) targetFound = true;
            // reflector check
            Actor reflector = getOneObjectAtOffset(0, 0, Reflector.class);
            if (reflector != null && reflector != source)
            {
                targetFound = true;
                target = reflector;
                // add reflected beam
                getWorld().addObject(new Beam(this, reflector), getX(), getY());
            }
            toMiddle(); // position back to current location
        }
    }
    
    /** internal method to position beam at current place in world */
    private void toMiddle()
    {
        setLocation(origX, origY);
        move((maxWide+shortened)/2);
    }

    /** internal method to place at current end-point of beam (not permanently) */
    private void toEnd()
    {
        setLocation(origX, origY);
        move(maxWide);
    }

    /**
     * used mainly for beams that no longer have a source;
     * also, called when its found target has changed
     * 
     * @return true if the actor is still in the world
     */
    private boolean shorten()
    {
        if (getImage().getWidth() > 2)
        {
            getImage().scale(getImage().getWidth()-2, 2);
            shortened += 2;
            toMiddle();
            return true;
        }
        if (getWorld() != null) getWorld().removeObject(this);
        return false;
    }
    
    /** internal method for beams that have not reached any target */
    private void lengthen()
    {
        maxWide += 2;
        getImage().scale(getImage().getWidth()+2, 2);
        toMiddle();
    }

    /**
     * removes the source object from a beam, if the source of the beam is the given actor;
     * if this is a reflected beam, its incident beam is slightly shortened and it target is
     * removed to allow it to create a new reflected beam
     * 
     * @param actor the object that must be the source of this beam for it to lose its source
     */
    public void cutOffSource(Actor actor)
    {
        // source check
        if (actor == source)
        {
            source = null; // lose source
            // incident beam check
            if (priorBeam != null)
            {
                // lose target
                priorBeam.targetFound = false;
                priorBeam.target = null;
                // slightly shorten length (allows for rotation of reflector into beam)
                for (int i=0; i<2; i++)
                {
                    priorBeam.maxWide -= 2;
                    priorBeam.shortened -= 2;
                    priorBeam.shorten();
                }
            }
        }
    }
}